package vishnu.testvis.object;

import vishnu.datablock.*;
import vishnu.testvis.visual.*;
import vishnu.testvis.dendro.*;
import java.util.StringTokenizer;
import java.util.Vector;
import java.io.*;
import java.net.*;

import vishnu.testvis.sammon.*;
import vishnu.util.Sammon;
import vishnu.cluster.Clustering;
import vishnu.testvis.browser.*;

public class DataManager
{
    private Vishnu displayFrame;
    private DataBlock data = null;
    private CacheManager cache = null;
    public Vector collections;
    public Collection currentCollection = null;
    public String currentView = null;
    /** Labels for the clusters needed by tree map,
     *  generated in this class */
    private String [] clusterLabels = null;
    /** vertical array of vectors representing
     *  the documents that word x appears in */
    private int [] clustersBigWord;
    //private Vector [] wordDocuments;
    /** ClusterObjs as used in Tree map */
    private DocObj [] docObjs;
    private ClusterObj [] clusterObjs = null;
    private ClusterGroupObj [] clusterGroups = null;
    private int[][] sommClusterWords;
    private Vector sammonClusters;

    private String engine_type = "Engine";
    
    /** Set gsdl state */
    public void setEngineType(String engine) {
	engine_type = engine;
    }
    /** Generate labels for the clusters needed by tree map,
     *  pre: data is initialised
     */

    private void setUpClusterLabels()
    {
        int [] wordFreqs = new int [data.words.length];
        StringBuffer sbuff= new StringBuffer(20);
        clustersBigWord = new int[data.clusters.length];
        clusterLabels= new String[data.clusters.length];
        sommClusterWords= new int [data.clusters.length][4];

        for(int c=0;c<data.clusters.length;c++)
        {
            for(int d=0; d < data.words.length; d++) wordFreqs[d]=0;

            for(int d=0; d < data.clusters[c].size(); d++)
            {
                Integer n = (Integer) data.clusters[c].elementAt(d);
                int docNo = n.intValue();
                SparseNode row = data.matrix.getRow(docNo);
                while( row!= null)
                {
                    wordFreqs[row.getColumn()]++;
                    row = row.getNext();
                }
            }

            int col1 = -1, c1val = 0;
            int col2 = -1, c2val = 0;
            int col3 = -1, c3val = 0;
            int col4 = -1, c4val = 0;
            for(int d=0; d < data.words.length; d++)
            {
                if(c1val < wordFreqs[d])
                {
                    col4=col3;
                    col3=col2;
                    c3val=c2val;
                    col2=col1;
                    c2val=c1val;
                    col1=d;
                    c1val=wordFreqs[d];
                }
                else if(c2val < wordFreqs[d])
                {
                    col4=col3;
                    col3=col2;
                    c3val=c2val;
                    col2=d;
                    c2val=wordFreqs[d];
                }
                else if(c3val < wordFreqs[d])
                {
                    col4=col3;
                    col3=d;
                    c3val=wordFreqs[d];
                }
                else if(c4val < wordFreqs[d])
                {
                    col4=d;
                    c4val=wordFreqs[d];
                }

            } // for (each document in the cluster)
           if(col1 > -1)
           {
                sbuff.append(data.words[col1]);
                if(col2 > -1)
                {
                    sbuff.append(" ");
                    sbuff.append(data.words[col2]);
                    if(col3 > -1)
                    {
                        sbuff.append(" ");
                        sbuff.append(data.words[col3]);
                    }  // if
                } // if
                clusterLabels[c] = new String(sbuff);
                sbuff.setLength(0);
            } //if
            else
                clusterLabels[c] = new String("<empty>");

            clustersBigWord[c] = col1;
            sommClusterWords[c][0] = col1;
            sommClusterWords[c][1] = col2;
            sommClusterWords[c][2] = col3;
            sommClusterWords[c][3] = col4;
        } // for (each cluster)
    } // end of setUpClusterLabels()
    

    private void setUpDocObjs()
    {
        docObjs = new DocObj[data.docs.length];
        for(int c=0;c<data.docs.length;c++)
            docObjs[c] = new DocObj(c, data.descriptions[c], data.matrix.getRow(c), data.docs[c]);
    }
    
    public DocObj getDocObjAt(int index)
    {
        return docObjs[index];
    }

    public int getClusterGroupsSize()
    {
        return clusterGroups.length;
    }

    public ClusterGroupObj getClusterGroup(int i)
    {
        return clusterGroups[i];
    }
    private void setUpClusterObjs()
    {
        clusterObjs = new ClusterObj[data.clusters.length];
        for(int c=0;c<data.clusters.length;c++)
        {
            DocObj [] dObjArray = new DocObj[data.clusters[c].size()];
            for(int d=0; d<data.clusters[c].size();d++)
            {
                Integer n = (Integer)data.clusters[c].elementAt(d);
                dObjArray[d] = docObjs[n.intValue()];
            }
            clusterObjs[c]= new ClusterObj(dObjArray,clusterLabels[c],data.sammon,this);
        }
        //ClusterObj(DocObj[] d, String t, Point2D sammon, DataManager dm)
        // !!!!remember to comment out this loop when it works
        //for(int c=0;c<data.clusters.length;c++)
        //    System.out.println("In cluster " + c  + clusterObjs[c].getDocsSize() + " documents ");
    }
    /** This for tree map as it needs clusters of clusters
     * this is a cheap and probably not very satisfactory solution
     * as all I am doing is taking clusters with the common most significant
     * word and putting them in groups.
     */
    private void setUpClusterGroupObjs()
    {
        int count = data.clusters.length;
        int num=0, i,j;
        int [] sizes = new int[count];
        for(i=0;i < count;i++)
            sizes[i]=1;

        // contains #supercluster clusterList objects, each a list of clusters
        Vector groupList = new Vector(count);
        
        // contains #supercluster BigWords for each supercluster
        Vector wordList = new Vector(count);
        for(i=0;i < count;i++)
        {
            if(sizes[i]>0)
            {
                Vector clusterList = new Vector(count);
                clusterList.addElement(clusterObjs[i]);
                int bigWord = clustersBigWord[i];
                for (j = i+1; j < count; j++)
                    if (clustersBigWord[j] == bigWord)
                    {
                        sizes[i]+=sizes[j];
                        if(sizes[j]>0)
                            clusterList.addElement(clusterObjs[j]);
                        sizes[j]=0;
                    }
                groupList.addElement(clusterList);
                wordList.addElement(new Integer(bigWord));
            }
        }

        clusterGroups = new ClusterGroupObj[groupList.size()];

        for(i=0;i < groupList.size();i++)
        {
            Vector clusterList = (Vector)groupList.elementAt(i);
            if(clusterList == null)
                System.out.println("Cluster list "+ i + " is null");
            else
            {
                //ClusterObj [] clusterObjsArray = clusterList.toArray();
                int groupSize = clusterList.size();
                ClusterObj [] clusterObjsArray = new ClusterObj [groupSize];
                for (j=0; j < groupSize;j++)
                    clusterObjsArray[j] = (ClusterObj) clusterList.elementAt(j);
                //System.out.println("Group " + i + " is " + clusterObjsArray.length + " long");
                Integer bigWord = (Integer)wordList.elementAt(i);
                int bw = bigWord.intValue();
                String word;
                if (bw <0)
                {
                    System.out.println("Group " + i + " is " + clusterObjsArray.length + " long");
                    System.out.println("bigword is empty");
                    word = "< Empty >";
                }
                else
                    word = data.words[bigWord.intValue()];
                clusterGroups[i] = new ClusterGroupObj(clusterObjsArray,word);
            }
        }
    }
    
    
    private void printData(String[] str, String filename)
    {
        try
	    {
	        filename += ".txt";
            File target = new File(filename);
		    FileWriter fw = new FileWriter(target);
		    PrintWriter out = new PrintWriter(fw,true);
		    String line = new String();
		   			
			    for( int j = 0; j < str.length; j++ )
			    {
			        line = str[j] + " ";
			        out.println(j + " " + line);
			        System.out.println(j + " " + line);
		        }
		    
		    out.close();
	    } catch (IOException e){ System.out.println("Error " + e.toString());
	    } catch (SecurityException se) { System.out.println("Error " + 	se.toString());}
	}
	
	public void setUpDendroClusters()
	{
	    displayFrame.dendroPanel.treePanel.getData();
	}
    
       
    
    public int query(String queryStr)
    {
        if( !displayFrame.refine )
        {
            if( displayFrame.firstSearch )
            {
                displayFrame.firstSearch = false;
            }
        }
                
	Engine theEngine;
	try {
	    theEngine = (Engine)Class.forName("vishnu.testvis.object."+engine_type).newInstance();
	} catch (Exception e) {
	    System.err.println("Couldn't create an engine "+engine_type+ ", aborting query");
	    System.err.println(e);
	    return 0;
	}
	theEngine.setDisplay(displayFrame);
	//GSDLEngine theEngine = new GSDLEngine(displayFrame);
	System.out.println("Searching for " + queryStr);
	DataBlock db =  theEngine.getData(queryStr, currentCollection.dir, currentView);

        if( db == null ) return 0;

	data = db;
	
        int spaceIndex=0;
        int count = 0;
        int wordNum = 0;
        int clusterNum=0;
        StringBuffer qbuff = new StringBuffer();
            
        setUpClusterLabels();
        setUpDocObjs();
        setUpClusterObjs();
        setUpClusterGroupObjs();
        setUpSammClusters();
        if( !displayFrame.refine ) setUpDendroClusters();
        
	count = this.data.docs.length;
	wordNum = this.data.words.length;
	clusterNum = this.data.clusters.length;
	
        if( !displayFrame.refine )
	    {
		displayFrame.searchStarted = true;
	    }
        
	displayFrame.update();
        return count;
    }


    public void fetch(int index)
    {
        if(data==null) return ;
        if(index < 0 || index >= data.docs.length) return;
        int docNum = data.docs[index];
        fetchDocNo(docNum);
    }


    public void fetchDocNo(long docNum)
    {
        System.out.println("Fetch document: " + Long.toString(docNum));

	Engine theEngine;
	try {
	    theEngine = (Engine)Class.forName("vishnu.testvis.object."+engine_type).newInstance();
	} catch (Exception e) {
	    System.err.println("Couldn't create an engine "+engine_type+ ", aborting query");
	    System.err.println(e);
	    return;
	}
	theEngine.setDisplay(displayFrame);
	//GSDLEngine engine = new GSDLEngine(displayFrame);
	Vector document = theEngine.getDocument(Long.toString(docNum),currentCollection.dir, currentView);
	
	// Fetch theFetch = new Fetch(displayFrame);
        // document = theFetch.getDocument(docNum);
        
	Browser browser;
	if (document.size() > 1) {
	    String url = (String)document.elementAt(1);
	    
	    if( url.indexOf("http") != -1 )
		{
		    try{
			displayFrame.getAppletContext().showDocument(new URL(url),"_blank");
			// http://icarus.umkc.edu/cgi-bin/ptext?doc=perseus:text:1999.01.0154"),"_blank");
		    }catch(IOException ex){ex.printStackTrace();};
		}     
	    else browser = new Browser(document,docNum);
	} else { 
	    browser = new Browser(document,docNum);
	}
    }
    

    public String [] getClusterLabels()
    {
        return clusterLabels;
    }


    public String [] getDescriptions()
    {
        if(data != null)
            return data.descriptions;
        else return new String[0];
    }
    
    
    public String [] getTrecIDs()
    {
        if(data != null)
        {
            String trecID[] = new String[data.descriptions.length];
            for( int i = 0; i < data.descriptions.length; i++ )
            {
                StringTokenizer st = new StringTokenizer(data.descriptions[i]);
                trecID[i] = st.nextToken();             
            }
            return trecID;
        }
        else return new String[0];
    }
    
       
    public String getDescriptionStr(Vector subset)
    {
        String returnStr = new String("");
        for(int c = 0; c < subset.size();c++)
        {
            Integer n = (Integer)(subset.elementAt(c));
            returnStr = returnStr + data.descriptions[n.intValue()] + "\n";
        }
        return returnStr;
    }
    public int [] getDocs()
    {
        if(data != null)
            return data.docs;
        else return new int[0];
    }

    public String [] getWords()
    {
        if(data != null)
        {
            if(data.words==null)
            {
                System.out.println("No words in data!");
                return new String[0];
            }
            return data.words;
        }
        else return new String[0];
    }

    public SparseMatrix getMatrix()
    {
        if(data != null)
            return data.matrix;
        else return null;
    }

    public Vector [] getClusters()
    {
        if(data != null)
            return data.clusters;
        else return null;
    }
    public Point2D [] getSammon()
    {
        if(data != null)
            return data.sammon;
        else return new Point2D[0];
    }

    public DataManager(Vishnu d) {
	this(d, null);
    }
    public DataManager(Vishnu d, String engine)
    {
        displayFrame = d;
	if (engine != null && !engine.equals("")) {
	    engine_type = engine;
	}
	Engine theEngine;
	try {
	    theEngine = (Engine)Class.forName("vishnu.testvis.object."+engine_type).newInstance();
	} catch (Exception e) {
	    System.err.println("Couldn't create an engine "+engine_type+ ", aborting query");
	    System.err.println(e);
	    return;
	}

	theEngine.setDisplay(displayFrame);
	Vector q = new Vector();
	q.addElement("initialise");
	collections = theEngine.getCollections(q); 
    }


    // This constructor creates a smaller data manager from
    // several clusters from a larger one
    // it is intended to be used with sammon mapping drilling down
    // The clustering code is lifted from the C version and could be
    // improved
    
    public DataManager(DataBlock oldData, Vector idVect, Vishnu d)
    {
        displayFrame = d;
        
        // create new document/keyword matrix with only the 
        // documents from the selected cluster, but keeping
        // the same 100 keywords as columns
        SparseMatrix m = new SparseMatrix(idVect.size(),
                                          oldData.matrix.getColumnCount());

        for (int c = 0; c< idVect.size();c++)
        {
          // the document id of the cth document
          Integer i = (Integer)idVect.elementAt(c);
          
          // ptr to the first document of ith row (linked list)
          SparseNode row = oldData.matrix.getRow(i.intValue());
          // add the row to new lean matrix
          while(row != null)
          {
              m.addDataItem(c,row.getColumn(),row.getValue());
              row=row.getNext();
          }
        }
        
        float [][] full = m.getExpanded(); 
        
        // the number of new clusters generated is either 20
        // or if there are not many docs half the #docs
        int clusterNum =  m.getRowCount()>20?10:m.getRowCount()/2;
        
        // limit set to 20 if > 20 docs otherwise to 3/4 #docs
        int limit =  m.getRowCount()>20?20:(m.getRowCount()/2)+
                      (m.getRowCount()/4);
                      
        Clustering clustering = new Clustering(Clustering.COMPLETE_LINKAGE,
					       full,/*m.getRowCount(),
						      m.getColumnCount(),*/1,
					       clusterNum,limit);
                                                    
        Vector [] clusters = clustering.getClusters();
        double [][] centroids= clustering.getCentroids();

        Sammon map = new Sammon(centroids);
        Point2D [] sammon = map.getMapping();
        String [] words = oldData.words;
        int [] newDocs = new int[idVect.size()];
        String [] newDesc = new String[idVect.size()];
        for (int c=0;c<idVect.size();c++)
        {
          Integer i = (Integer) idVect.elementAt(c);
          int index = i.intValue();
          newDocs[c] = oldData.docs[index];
          newDesc[c] = oldData.descriptions[index];
        }
        data = new DataBlock(newDocs,newDesc,words,m,clusters,sammon);
        
        setUpClusterLabels();
        setUpDocObjs();
        setUpClusterObjs();
        setUpClusterGroupObjs();
        setUpSammClusters();
    }

    public DataBlock getDataBlock()
    {
        return data;
    }
    
    public int getWordCount()
    {
        if (data!=null)
            return data.words.length;
        else return 0;
    }
    public int getDocCount()
    {
        if (data!=null)
            return data.docs.length;
        return 0;
    }

    public Vector getSammonClusters()
    {
        if(data != null)
            return sammonClusters;
        else return null;
    }

    private void setUpSammClusters()
    {
        LengthTable [] table = new LengthTable [data.clusters.length];
        
        // create as many sammon clusters as there are clusters in the data object
        // and store it in sammmonClusters
        sammonClusters = new Vector(data.clusters.length);
        
        for (int c = 0; c < data.clusters.length; c++){
	    table[c]= new LengthTable(c,data.clusters[c].size());
        }
        
        
        // sort by cluster size, I guess
        QSortAlgorithmTableLength qSort = new QSortAlgorithmTableLength();
        try
        {
            qSort.sort(table);
        }
        catch (Exception e)
        {}
        
        
        for (int c = 0; c < data.clusters.length; c++)
        {
            int i = table[c].index;
             SammCluster cluster = new SammCluster(data.clusters[i],
                        data.sammon[i], sommClusterWords[i],
                        clusterObjs[i]);
            sammonClusters.addElement(cluster);
        }
    }

    public void requestCollection()
    {
	// should reset the data
	data = null;
    }
}













